#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.
# Copyright 2019 The logging-gelf Authors. All rights reserved.

import logging

SYSLOG_LEVELS = {
    logging.CRITICAL: 2, logging.ERROR: 3, logging.WARNING: 4, logging.INFO: 6,
    logging.DEBUG: 7,
}


# monkey patch
def process(self, msg, kwargs):
    """
    Process the logging message and keyword arguments passed in to
    a logging call to insert contextual information. You can either
    manipulate the message itself, the keyword args or both. Return
    the message and kwargs modified (or not) to suit your needs.

    Normally, you'll only need to override this one method in a
    LoggerAdapter subclass for your specific needs.
    """
    extra = dict(self.extra)
    extra.update(kwargs.get('extra', dict()))
    kwargs["extra"] = extra
    return msg, kwargs


# noinspection PyPep8
from logging import LoggerAdapter

LoggerAdapter.process = process



import unittest
import logging
import io
import json

from logging_gelf.formatters import GELFFormatter


class TestFormatter(unittest.TestCase):
    def setUp(self) -> None:
        self.logger = logging.getLogger("gelf")
        self.logger.setLevel(logging.INFO)
        self.log_stream = io.StringIO()

        handler = logging.StreamHandler(self.log_stream)
        handler.setFormatter(GELFFormatter())
        self.logger.addHandler(handler)

        super().setUp()

    def test_handle_format_exception(self):
        self.logger.info("hello %s!", "beautiful world!")

        log = json.loads(self.log_stream.getvalue())
        self.assertNotEqual(log, "")

    def test_has_host(self):
        self.logger.info("hello %s!", "world")

        log = json.loads(self.log_stream.getvalue())
        self.assertIn("host", log)
        self.assertNotEqual("", log["host"])

    def test_short_message(self):
        self.logger.info("hello %s!", "world")

        log = json.loads(self.log_stream.getvalue())
        self.assertEqual("hello world!", log["short_message"])

    def test_exception(self):
        try:
            raise Exception("failed to xxx")
        except Exception:
            self.logger.exception("hello!")

        log = json.loads(self.log_stream.getvalue())
        self.assertIn("failed to xxx", log["full_message"])
        self.assertIn("hello!", log["full_message"])

    def test_no_exception(self):
        self.logger.info("hello!")

        log = json.loads(self.log_stream.getvalue())
        self.assertNotIn("full_message", log)

    def test_stack_info(self):
        self.logger.info("hello!", stack_info=True)

        log = json.loads(self.log_stream.getvalue())
        self.assertIn("full_message", log)

    def test_no_stack_info(self):
        self.logger.info("hello!")

        log = json.loads(self.log_stream.getvalue())
        self.assertNotIn("full_message", log)

    def test_exception_as_message(self):
        try:
            raise Exception("failed to xxx")
        except Exception as exc:
            self.logger.exception(exc)

        log = json.loads(self.log_stream.getvalue())
        self.assertEqual("failed to xxx", log["short_message"])



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.
# Copyright 2019 The logging-gelf Authors. All rights reserved.

"""
Formatters specify the layout of log records in the final output (GELF).
"""

import json
import logging
import re

from logging_gelf.schemas import GelfSchema


class StringJSONEncoder(json.JSONEncoder):
    def encode(self, o):
        """description of encode"""
        res = dict()
        for key, value in o.items():
            if isinstance(value, (list, tuple, dict)):
                res[key] = str(value)
            else:
                res[key] = value
        return json.JSONEncoder.encode(self, res)


class GELFFormatter(logging.Formatter):
    """A GELF formatter to format a :class:`logging.LogRecord` into GELF.

    :param logging_gelf.schemas.GelfSchema schema: The marshmallow schema to
    use to format data.
    :param bool null_character: Append a '\0' at the end of the string. It
    depends on the input used.
    :param list|None exclude_patterns: List of regexp used to exclude keys.
    """

    def __init__(self, schema=GelfSchema, null_character=False,
                 JSONEncoder=json.JSONEncoder, exclude_patterns=None):
        if not issubclass(schema, GelfSchema):
            raise ValueError("Schema MUST inherit from 'GelfSchema'")

        self.schema = schema
        self.null_character = null_character
        self._encoder_cls = JSONEncoder
        exclude_patterns = exclude_patterns if exclude_patterns else list()
        self.exclude_patterns = ['^_{}'.format(x) for x in exclude_patterns]
        logging.Formatter.__init__(self)

    @staticmethod
    def get_record_message(record):
        """Get record message

        :param logging.LogRecord record: Contains all the information pertinent
        to the event being logged.
        :return: Formatted message.
        :rtype: str
        """
        # noinspection PyBroadException
        try:
            return record.getMessage() % vars(record)
        except Exception:
            return record.getMessage()

    def serialize_record(self, record):
        """Serialize logging record into a dict

        :param logging.LogRecord record: Contains all the information pertinent
        to the event being logged.
        :return: A dict dump of the record.
        :rtype: dict
        """
        # exc_info, exc_text and stack_info logic stolen from the standard library
        # https://github.com/python/cpython/blob/3.8/Lib/logging/__init__.py#L655

        if record.exc_info:
            # Cache the traceback text to avoid converting it multiple times
            # (it's constant anyway)
            if not record.exc_text:
                record.exc_text = self.formatException(record.exc_info)

        record.message = self.get_record_message(record)
        record.full_message = ""

        if record.exc_text:
            if record.message[-1:] != "\n":
                record.full_message = record.full_message + "\n"
            record.full_message = record.full_message + record.exc_text

        if record.stack_info:
            if record.full_message[-1:] != "\n":
                record.full_message = record.full_message + "\n"
            record.full_message = record.full_message + self.formatStack(
                record.stack_info)

        if record.full_message != "":
            record.full_message = record.message + "\n" + record.full_message

        return self.filter_keys(self.schema().dump(record))

    def format(self, record):
        """Format the specified record into json using the schema which MUST
        inherit from :class:`logging_gelf.schemas.GelfSchema`.

        :param logging.LogRecord record: Contains all the information pertinent
        to the event being logged.
        :return: A JSON dump of the record.
        :rtype: str
        """
        out = json.dumps(self.serialize_record(record), cls=self._encoder_cls)
        if self.null_character is True:
            out += '\0'
        return out

    def filter_keys(self, data):
        """Filter GELF record keys using exclude_patterns

        :param dict data: Log record has dict
        :return: the filtered log record
        :rtype: dict
        """
        if self.exclude_patterns is None:
            return data

        keys = list(data.keys())
        for pattern in self.exclude_patterns:
            for key in keys:
                if re.match(pattern, key):
                    keys.remove(key)

        return dict(filter(lambda x: x[0] in keys, data.items()))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.
# Copyright 2019 The logging-gelf Authors. All rights reserved.

"""
Schema allow to specify a mapping for :class:`logging.LogRecord`. It based on
:class:`marshmallow.Schema`. All schema MUST inherit from
:class:`logging_gelf.schemas.GelfSchema`.
"""
import socket
import time

from marshmallow import Schema, fields, EXCLUDE, post_dump
from logging_gelf import SYSLOG_LEVELS


GELF_1_1_FIELDS = [
    'version', 'host', 'short_message', 'full_message', 'timestamp', 'level',
    'line', 'file'
]


class GelfSchema(Schema):
    class Meta:
        unknown = EXCLUDE

    version = fields.Constant("1.1")
    host = fields.Constant(socket.gethostname())
    message = fields.String(data_key="short_message")
    full_message = fields.String()
    timestamp = fields.Method('to_timestamp')
    level = fields.Method('to_syslog_level')
    lineno = fields.Integer(data_key="line")
    pathname = fields.String(data_key="file")

    @classmethod
    def _forge_key(cls, key, _value):
        return key

    @classmethod
    def to_syslog_level(cls, value):
        """description of to_syslog_level"""
        return SYSLOG_LEVELS.get(value.levelno, 1)

    @classmethod
    def to_timestamp(cls, value):
        """to_timestamp"""
        if value.created:
            return value.created
        else:
            return time.time()

    @classmethod
    def format_key(cls, xpath, key, value):
        if xpath in (None, ""):
            if key in GELF_1_1_FIELDS:
                return key
            elif key in (None, ""):
                return ""
            else:
                return "_{}".format(cls._forge_key(key, value))
        else:
            return "{}_{}".format(xpath, cls._forge_key(key, value))

    @classmethod
    def to_flat_dict(cls, xpath, key, value):
        parts = dict()
        if isinstance(value, dict):
            for subkey, subvalue in value.items():
                parts.update(cls.to_flat_dict(
                    cls.format_key(xpath, key, value), subkey, subvalue
                ))
        elif isinstance(value, (list, tuple)):
            if len(value) < 20:
                for idx in range(len(value)):
                    parts.update(cls.to_flat_dict(
                        cls.format_key(xpath, key, value), idx, value[idx]
                    ))
            else:
                try:
                    parts[cls.format_key(xpath, key, value)] = str(value)
                except Exception:
                    pass
        else:
            parts[cls.format_key(xpath, key, value)] = value
        return parts

    @post_dump
    def fix_additional_fields(self, data, **kwargs):
        """description of fix_additional_fields"""
        return self.to_flat_dict("", "", data)

    @post_dump
    def remove_empty_values(self, data, **kwargs):
        return {
            key: value for key, value in data.items()
            if value is not None and value != ""
        }
